import hashlib
import logging
import re
from datetime import datetime, timedelta
from io import StringIO
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, Tuple, Union, cast

import pandas as pd
import paramiko
import requests

from ...core import Flight, Traffic
from ...core.time import round_time, split_times, timelike, to_datetime
from ..basic.airport import Airport


class OpenSky(object):

    columns = ["icao24", "callsign", "origin_country", "last_position",
               "timestamp", "longitude", "latitude", "altitude", "onground",
               "ground_speed", "track", "vertical_rate", "sensors",
               "baro_altitude", "squawk", "spi", "position_source"]

    basic_request = ("select * from state_vectors_data4 {other_columns} "
                     "where hour>={before_hour} and hour<{after_hour} "
                     "and time>={before_time} and time<{after_time} "
                     "{other_where}")

    airport_request = ("select icao24, callsign from state_vectors_data4 "
                       "where lat<={airport_latmax} and lat>={airport_latmin} "
                       "and lon<={airport_lonmax} and lon>={airport_lonmin} "
                       "and baroaltitude<=1000 "
                       "and hour>={before_hour} and hour<{after_hour} "
                       "and time>={before_time} and time<{after_time} "
                       "group by icao24, callsign")

    sensor_request = ("select icao24, callsign, s.ITEM as serial,"
                      "count(*) as count from "
                      "state_vectors_data4, state_vectors_data4.serials s "
                      "where hour>={before_hour} and hour<{after_hour} "
                      "{other_where} group by icao24, callsign, s.ITEM")

    shell: Optional[Any]  # the connection to the Impala shell

    def __init__(self, username: str, password: str, cache_dir: Path) -> None:

        self.username = username
        self.password = password
        self.connected = False
        self.cache_dir = cache_dir
        self.shell = None
        if not self.cache_dir.exists():
            self.cache_dir.mkdir(parents=True)

        if username == "" or password == "":
            self.auth = None
        else:
            self.auth = (username, password)

    def clear_cache(self) -> None:
        for file in self.cache_dir.glob('*'):
            file.unlink()

    @staticmethod
    def _format_dataframe(df: pd.DataFrame,
                          nautical_units=True) -> pd.DataFrame:

        df.callsign = df.callsign.str.strip()

        if nautical_units:
            df.altitude = df.altitude / 0.3048
            df.baro_altitude = df.baro_altitude / 0.3048
            df.ground_speed = df.ground_speed / 1852 * 3600
            df.vertical_rate = df.vertical_rate / 0.3048 * 60

        df.timestamp = df.timestamp.apply(datetime.fromtimestamp)
        # warning is raised here: SettingWithCopyWarning:
        # A value is trying to be set on a copy of a slice from a DataFrame.
        # Try using .loc[row_indexer,col_indexer] = value instead
        pd.options.mode.chained_assignment = None
        df = df.loc[df.last_position.notnull()]  # do we really miss much?
        df.last_position = df.last_position.apply(datetime.fromtimestamp)
        df = df.sort_values('timestamp')

        return df

    def _connect(self) -> None:
        if self.username == "" or self.password == "":
            raise RuntimeError("This method requires authentication.")
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect("data.opensky-network.org", port=2230,
                       username=self.username, password=self.password)
        self.shell = client.invoke_shell()
        self.connected = True
        total = ""
        while len(total) == 0 or total[-19:] != "[hadoop-1:21000] > ":
            b = self.shell.recv(256)
            total += b.decode()

    def _impala(self, request: str) -> Optional[pd.DataFrame]:

        digest = hashlib.md5(request.encode('utf8')).hexdigest()
        cachename = self.cache_dir / digest

        if not cachename.exists():
            if not self.connected:
                self._connect()

            # Now we are connected so self.c is not None
            self.shell = cast(Any, self.shell)

            logging.info("Sending request: {}".format(request))
            self.shell.send(request + ";\n")
            total = ""
            while len(total) == 0 or total[-19:] != "[hadoop-1:21000] > ":
                b = self.shell.recv(256)
                total += b.decode()
            with cachename.open('w') as fh:
                fh.write(total)

        logging.info("Reading request in cache {}".format(cachename))
        with cachename.open('r') as fh:
            s = StringIO()
            count = 0
            for line in fh.readlines():
                if re.match("\|.*\|", line):
                    count += 1
                    s.write(re.sub(" *\| *", ",", line)[1:-2])
                    s.write("\n")
            if count > 0:
                s.seek(0)
                df = pd.read_csv(s)
                return df

        return None

    def history(self, before: timelike, after: Optional[timelike]=None,
                date_delta: timedelta=timedelta(hours=1),
                callsign: Optional[Union[str, Iterable[str]]]=None,
                serials: Optional[Iterable[int]]=None,
                bounds: Optional[Tuple[float, float, float, float]]=None,
                other_columns: str="", other_where: str="",
                progressbar: Callable[[Iterable], Iterable]=iter
                ) -> Optional[Union[Traffic, Flight]]:

        return_flight = False
        before = to_datetime(before)
        if after is not None:
            after = to_datetime(after)
        else:
            after = before + timedelta(days=1)

        if isinstance(serials, Iterable):
            other_columns += ", state_vectors_data4.serials s "
            other_where += "and s.ITEM in {} ".format(tuple(serials))

        if isinstance(callsign, str):
            other_where += "and callsign='{:<8s}' ".format(callsign)
            return_flight = True

        elif isinstance(callsign, Iterable):
            callsign = ",".join("'{:<8s}'".format(c) for c in callsign)
            other_where += "and callsign in ({}) ".format(callsign)

        if bounds is not None:
            try:
                # thinking of shapely bounds attribute (in this order)
                # I just don't want to add the shapely dependency here
                west, south, east, north = bounds.bounds  # type: ignore
            except AttributeError:
                west, south, east, north = bounds

            other_where += "and lon>={} and lon<={} ".format(west, east)
            other_where += "and lat>={} and lat<={} ".format(south, north)

        cumul = []
        sequence = list(split_times(before, after, date_delta))

        for bt, at, bh, ah in progressbar(sequence):

            logging.info(
                f"Sending request between time {bt} and {at} "
                f"and hour {bh} and {ah}")

            request = self.basic_request.format(
                before_time=bt.timestamp(), after_time=at.timestamp(),
                before_hour=bh.timestamp(), after_hour=ah.timestamp(),
                other_columns=other_columns, other_where=other_where)

            df = self._impala(request)

            if df is None:
                continue

            df = df.drop(['lastcontact', ], axis=1)

            # TODO option for callsign/icao24
            # TODO remove serials as well (option)
            df = df.drop_duplicates(('callsign', 'time'))
            if df.lat.dtype == object:
                df = df[df.lat != 'lat']  # header is regularly repeated

            # restore all types
            for column_name in ['lat', 'lon', 'velocity', 'heading',
                                'geoaltitude', 'baroaltitude', 'vertrate',
                                'time', 'lastposupdate']:
                df[column_name] = df[column_name].astype(float)

            df.icao24 = df.icao24.apply(
                lambda x: "{:0>6}".format(hex(int(str(x), 16))[2:]))

            if df.onground.dtype != bool:
                df.onground = (df.onground == 'true')
                df.alert = (df.alert == 'true')
                df.spi = (df.spi == 'true')

            # better (to me) formalism about columns
            df = df.rename(columns={'lat': 'latitude',
                                    'lon': 'longitude',
                                    'heading': 'track',
                                    'velocity': 'ground_speed',
                                    'vertrate': 'vertical_rate',
                                    'baroaltitude': 'baro_altitude',
                                    'geoaltitude': 'altitude',
                                    'time': 'timestamp',
                                    'lastposupdate': 'last_position'})

            df = self._format_dataframe(df)
            cumul.append(df)

        if len(cumul) == 0:
            return None

        if return_flight:
            return Flight(pd.concat(cumul))

        return Traffic(pd.concat(cumul))

    def sensors(
            self, before: timelike, after: timelike,
            bounds: Tuple[float, float, float, float]
    ) -> Optional[pd.DataFrame]:

        before_hour = round_time(before, "before")
        after_hour = round_time(after, "after")

        try:
            # thinking of shapely bounds attribute (in this order)
            # I just don't want to add the shapely dependency here
            west, south, east, north = bounds.bounds  # type: ignore
        except AttributeError:
            west, south, east, north = bounds

        other_where = "and lon>={} and lon<={} ".format(west, east)
        other_where += "and lat>={} and lat<={} ".format(south, north)

        query = self.sensor_request.format(before_hour=before_hour,
                                           after_hour=after_hour,
                                           other_where=other_where)

        logging.info(f"Sending request: {query}")
        df = self._impala(query)
        if df is None:
            return None

        df = df[df['count'] != 'count']
        df['count'] = df['count'].astype(int)

        return df

    def online_aircraft(self, own=False) -> pd.DataFrame:
        what = "own" if (own and self.username != ""
                         and self.password != "") else "all"
        c = requests.get(f"https://opensky-network.org/api/states/{what}",
                         auth=self.auth)
        if c.status_code != 200:
            raise ValueError(c.content.decode())
        r = pd.DataFrame.from_records(c.json()['states'], columns=self.columns)
        r = r.drop(['origin_country', 'spi', 'sensors'], axis=1)
        r = r.dropna()
        return self._format_dataframe(r, nautical_units=True)

    def at_airport(self, before: timelike, after: timelike,
                   airport: Airport) -> Optional[pd.DataFrame]:

        before = to_datetime(before)
        after = to_datetime(after)

        before_hour = round_time(before, how='before')
        after_hour = round_time(after, how='after')

        request = self.airport_request.format(
            before_time=before.timestamp(),
            after_time=after.timestamp(),
            before_hour=before_hour.timestamp(),
            after_hour=after_hour.timestamp(),
            airport_latmax=airport.lat + 0.1,
            airport_latmin=airport.lat - 0.1,
            airport_lonmax=airport.lon + 0.1,
            airport_lonmin=airport.lon - 0.1,)

        df = self._impala(request)

        return df



import operator
import pickle
import re
import warnings
import zipfile
from functools import lru_cache
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple
from xml.etree import ElementTree

from shapely.geometry import Polygon
from shapely.ops import cascaded_union

from ...core.sector import components  # to be moved here TODO
from ...core.sector import (
    ExtrudedPolygon,
    Sector,
    SectorInfo,
    SectorList,
    cascaded_union_with_alt,
)


class SectorParser(object):

    ns = {
        "adrmsg": "http://www.eurocontrol.int/cfmu/b2b/ADRMessage",
        "aixm": "http://www.aixm.aero/schema/5.1",
        "gml": "http://www.opengis.net/gml/3.2",
        "xlink": "http://www.w3.org/1999/xlink",
    }

    airac_path: Optional[Path] = None
    cache_dir: Optional[Path] = None

    def __init__(self, config_file: Path) -> None:
        self.config_file = config_file
        self.initialized = False

    def init_cache(self) -> None:

        msg = f"Edit file {self.config_file} with AIRAC directory"

        if self.airac_path is None or self.cache_dir is None:
            raise RuntimeError(msg)

        self.full_dict: Dict[str, Any] = {}
        self.all_points: Dict[str, Tuple[float, ...]] = {}

        assert self.airac_path.is_dir()

        cache_file = self.cache_dir / "airac.cache"
        if cache_file.exists():
            with cache_file.open("rb") as fh:
                self.full_dict, self.all_points, self.tree = pickle.load(fh)
                self.initialized = True
                return

        for filename in [
            "Airspace.BASELINE",
            "DesignatedPoint.BASELINE",
            "Navaid.BASELINE",
        ]:

            if ~(self.airac_path / filename).exists():
                zippath = zipfile.ZipFile(
                    self.airac_path.joinpath(f"{filename}.zip").as_posix()
                )
                zippath.extractall(self.airac_path.as_posix())

        self.tree = ElementTree.parse(
            (self.airac_path / "Airspace.BASELINE").as_posix()
        )

        for airspace in self.tree.findall(
            "adrmsg:hasMember/aixm:Airspace", self.ns
        ):

            identifier = airspace.find("gml:identifier", self.ns)
            assert identifier is not None
            assert identifier.text is not None
            self.full_dict[identifier.text] = airspace

        points = ElementTree.parse(
            (self.airac_path / "DesignatedPoint.BASELINE").as_posix()
        )

        for point in points.findall(
            "adrmsg:hasMember/aixm:DesignatedPoint", self.ns
        ):

            identifier = point.find("gml:identifier", self.ns)
            assert identifier is not None
            assert identifier.text is not None

            floats = point.find(
                "aixm:timeSlice/aixm:DesignatedPointTimeSlice/"
                "aixm:location/aixm:Point/gml:pos",
                self.ns,
            )
            assert floats is not None
            assert floats.text is not None

            self.all_points[identifier.text] = tuple(
                float(x) for x in floats.text.split()
            )

        points = ElementTree.parse(
            (self.airac_path / "Navaid.BASELINE").as_posix()
        )

        for point in points.findall("adrmsg:hasMember/aixm:Navaid", self.ns):

            identifier = point.find("gml:identifier", self.ns)
            assert identifier is not None
            assert identifier.text is not None

            floats = point.find(
                "aixm:timeSlice/aixm:NavaidTimeSlice/"
                "aixm:location/aixm:ElevatedPoint/gml:pos",
                self.ns,
            )
            assert floats is not None
            assert floats.text is not None

            self.all_points[identifier.text] = tuple(
                float(x) for x in floats.text.split()
            )

        with cache_file.open("wb") as fh:
            pickle.dump((self.full_dict, self.all_points, self.tree), fh)

        self.initialized = True

    def append_coords(self, lr, block_poly):
        coords: List[Tuple[float, ...]] = []
        gml, xlink = self.ns["gml"], self.ns["xlink"]
        for point in lr.iter():
            if point.tag in ("{%s}pos" % (gml), "{%s}pointProperty" % (gml)):
                if point.tag.endswith("pos"):
                    coords.append(tuple(float(x) for x in point.text.split()))
                else:
                    points = point.attrib["{%s}href" % (xlink)]
                    coords.append(self.all_points[points.split(":")[2]])
        block_poly.append(
            (Polygon([(lon, lat) for lat, lon in coords]), None, None)
        )

    @lru_cache(None)
    def make_polygon(self, airspace) -> SectorList:
        polygons: SectorList = []
        designator = airspace.find("aixm:designator", self.ns)
        if designator is not None:
            name = designator.text
        for block in airspace.findall(
            "aixm:geometryComponent/aixm:AirspaceGeometryComponent/"
            "aixm:theAirspaceVolume/aixm:AirspaceVolume",
            self.ns,
        ):
            block_poly: SectorList = []
            upper = block.find("aixm:upperLimit", self.ns)
            lower = block.find("aixm:lowerLimit", self.ns)

            upper = (
                float(upper.text)
                if upper is not None and re.match("\d{3}", upper.text)
                else float("inf")
            )
            lower = (
                float(lower.text)
                if lower is not None and re.match("\d{3}", lower.text)
                else float("-inf")
            )

            for component in block.findall(
                "aixm:contributorAirspace/aixm:AirspaceVolumeDependency/"
                "aixm:theAirspace",
                self.ns,
            ):
                key = component.attrib["{http://www.w3.org/1999/xlink}href"]
                key = key.split(":")[2]
                child = self.full_dict[key]
                for ats in child.findall(
                    "aixm:timeSlice/aixm:AirspaceTimeSlice", self.ns
                ):
                    new_d = ats.find("aixm:designator", self.ns)
                    new_t = ats.find("aixm:type", self.ns)
                    if new_d is not None:
                        if designator is not None:
                            components[name].add(
                                SectorInfo(
                                    new_d.text,
                                    new_t.text if new_t is not None else None,
                                )
                            )
                        block_poly += self.make_polygon(ats)
                    else:
                        for sub in ats.findall(
                            "aixm:geometryComponent/"
                            "aixm:AirspaceGeometryComponent/"
                            "aixm:theAirspaceVolume/aixm:AirspaceVolume",
                            self.ns,
                        ):

                            assert sub.find("aixm:lowerLimit", self.ns) is None

                            for lr in sub.findall(
                                "aixm:horizontalProjection/aixm:Surface/"
                                "gml:patches/gml:PolygonPatch/gml:exterior/"
                                "gml:LinearRing",
                                self.ns,
                            ):
                                self.append_coords(lr, block_poly)

                    break  # only one timeslice

            if upper == float("inf") and lower == float("-inf"):
                polygons += cascaded_union_with_alt(block_poly)
            else:
                polygons.append(
                    ExtrudedPolygon(
                        cascaded_union([p for (p, *_) in block_poly]),
                        lower,
                        upper,
                    )
                )

        return cascaded_union_with_alt(polygons)

    def __getitem__(self, name: str) -> Optional[Sector]:
        return next(self.search(name, operator.eq), None)

    def search(self, name: str, cmp: Callable = re.match) -> Iterator[Sector]:

        if not self.initialized:
            self.init_cache()

        polygon = None
        type_: Optional[str] = None

        names = name.split("/")
        if len(names) > 1:
            name, type_ = names

        for airspace in self.tree.findall(
            "adrmsg:hasMember/aixm:Airspace", self.ns
        ):
            for ts in airspace.findall(
                "aixm:timeSlice/aixm:AirspaceTimeSlice", self.ns
            ):

                designator = ts.find("aixm:designator", self.ns)

                if (
                    designator is not None
                    and cmp(name, designator.text)
                    and (
                        type_ is None
                        or ts.find("aixm:type", self.ns).text == type_
                    )
                ):

                    polygon = self.make_polygon(ts)
                    type_ = ts.find("aixm:type", self.ns).text
                    if len(polygon) > 0:
                        yield Sector(designator.text, polygon, type_)
                    else:
                        warnings.warn(
                            f"{designator.text} produces an empty sector",
                            RuntimeWarning,
                        )

    def parse(self, pattern: str, cmp=re.match):

        if not self.initialized:
            self.init_cache()

        name = pattern
        names = name.split("/")
        type_pattern: Optional[str] = None

        if len(names) > 1:
            name, type_pattern = names

        for airspace in self.tree.findall(
            "adrmsg:hasMember/aixm:Airspace", self.ns
        ):
            for ts in airspace.findall(
                "aixm:timeSlice/aixm:AirspaceTimeSlice", self.ns
            ):

                type_ = ts.find("aixm:type", self.ns)
                designator = ts.find("aixm:designator", self.ns)

                if (
                    type_pattern is None
                    or (type_ is not None and type_.text == type_pattern)
                ) and (designator is not None and cmp(name, designator.text)):
                    yield SectorInfo(
                        designator.text,
                        type_.text if type_ is not None else None,
                    )



import json
from pathlib import Path

from shapely.geometry import shape

from ...core.sector import ExtrudedPolygon, Sector

with Path(__file__).absolute().with_name("firs.json").open("r") as fh:
    fir = json.loads("".join(fh.readlines()))

eurofirs = {
    elt["properties"]["IDENT"]: Sector(
        name=elt["properties"]["NAME"][:-4],  # Remove " FIR" at the end
        elements=[
            ExtrudedPolygon(
                shape(elt["geometry"]),
                int(elt["properties"]["LOWERLIMIT"]),
                int(elt["properties"]["UPPERLIMIT"]),
            )
        ],
        type_=elt["properties"]["TYPE"],
    )
    for elt in fir["features"]
}



from collections import UserDict
from contextlib import contextmanager
from typing import Optional

import numpy as np

from fastkml import LineStyle, PolyStyle, kml
from fastkml.geometry import Geometry
from shapely.geometry import LineString

from ..core import Flight, Sector

_colormap = {
    "blue": "#1928f0",
    "red": "#ff3900",
    "green": "#19ff00",
    "yellow": "#effe00",
    "magenta": "#f300b0",
    "cyan": "#02d8ec",
    "orange": "#ffc000",
}

current_document: Optional[kml.Document] = None


class StyleMap(UserDict):
    def __missing__(self, color: str):
        style = kml.Style(id=color)
        if current_document is not None:
            current_document.append_style(style)
        style.append_style(LineStyle(color=color, width=2))
        style.append_style(PolyStyle(color=color))
        styleUrl = kml.StyleUrl(url=color)
        return styleUrl


def toStyle(color: str, alpha: float = 0.5) -> kml.StyleUrl:
    # saturate alpha
    alpha = max(0, alpha)
    alpha = min(1, alpha)
    color = _colormap.get(color, color).upper()
    if len(color) != 7:
        raise ValueError("Invalid color")
    color = color[1:]
    # KML colors are f***ed...
    key = hex(int(alpha * 255))[2:] + color[4:6] + color[2:4] + color[:2]
    return _stylemap[key]


@contextmanager
def export(filename: str):
    global current_document
    kml_tree = kml.KML()
    current_document = kml.Document()
    yield current_document
    kml_tree.append(current_document)
    with open(filename, "w", encoding="utf8") as kml_file:
        kml_file.write(kml_tree.to_string(prettyprint=True))
    _stylemap.clear()
    current_document = None


def _flight_export_kml(
    flight: Flight,
    styleUrl: Optional[kml.StyleUrl] = None,
    color: Optional[str] = None,
    alpha: float = .5,
    **kwargs,
) -> kml.Placemark:
    if color is not None:
        # the style will be set only if the kml.export context is open
        styleUrl = toStyle(color)
    params = {
        "name": flight.callsign,
        "description": flight.info_html(),
        "styleUrl": styleUrl,
    }
    for key, value in kwargs.items():
        params[key] = value
    placemark = kml.Placemark(**params)
    placemark.visibility = 1
    # Convert to meters
    coords = np.stack(flight.coords)
    coords[:, 2] *= 0.3048
    placemark.geometry = Geometry(
        geometry=LineString(coords),
        extrude=True,
        altitude_mode="relativeToGround",
    )
    return placemark


def _sector_export_kml(
    sector: Sector,
    styleUrl: Optional[kml.StyleUrl] = None,
    color: Optional[str] = None,
    alpha: float = .5,
) -> kml.Placemark:
    if color is not None:
        # the style will be set only if the kml.export context is open
        styleUrl = toStyle(color)
    folder = kml.Folder(name=sector.name, description=sector.type)
    for extr_p in sector:
        for elt in sector.decompose(extr_p):
            placemark = kml.Placemark(styleUrl=styleUrl)
            placemark.geometry = kml.Geometry(
                geometry=elt, altitude_mode="relativeToGround"
            )
            folder.append(placemark)
    return folder


setattr(Flight, "export_kml", _flight_export_kml)
setattr(Sector, "export_kml", _sector_export_kml)
_stylemap = StyleMap()



import logging
from functools import lru_cache, partial
from pathlib import Path
from typing import Optional, Tuple, Union

import pandas as pd
import pyproj
from shapely.ops import transform


class DataFrameMixin(object):
    def __init__(self, data: pd.DataFrame) -> None:
        self.data: pd.DataFrame = data

    @classmethod
    def from_file(
        cls, filename: Union[Path, str]
    ) -> Optional["DataFrameMixin"]:
        path = Path(filename)
        if path.suffixes == [".pkl"]:
            return cls(pd.read_pickle(path))
        if path.suffixes == [".csv"]:
            return cls(pd.read_csv(path))
        if path.suffixes == [".h5"]:
            return cls(pd.read_hdf(path))
        return None

    # --- Special methods ---

    def _repr_html_(self):
        return self.data._repr_html_()

    def __repr__(self):
        return self.data.__repr__()

    def __len__(self) -> int:
        return self.data.shape[0]

    # --- Redirected to pandas.DataFrame ---

    def to_pickle(self, filename: Union[str, Path]) -> None:
        self.data.to_pickle(filename)

    def to_csv(self, filename: Union[str, Path]) -> None:
        self.data.to_csv(filename)

    def to_hdf(self, filename: Union[str, Path]) -> None:
        self.data.to_hdf(filename)

    def to_json(self, filename: Union[str, Path]) -> None:
        self.data.to_json(filename)

    def to_excel(self, filename: Union[str, Path]) -> None:
        self.data.to_excel(filename)

    def query(self, query: str) -> "DataFrameMixin":
        return self.__class__(self.data.query(query))

    def groupby(self, *args, **kwargs):
        return self.data.groupby(*args, **kwargs)


class ShapelyMixin(object):

    # --- Properties ---

    @property
    def bounds(self) -> Tuple[float, float, float, float]:
        # just natural!
        return self.shape.bounds  # type: ignore

    @property
    def extent(self) -> Tuple[float, float, float, float]:
        # convenient for ax.set_extent
        west, south, east, north = self.bounds
        return west, east, south, north

    @property
    def centroid(self):
        return self.shape.centroid

    @property
    def area(self) -> float:
        return self.project_shape().area

    # --- Representations ---

    def _repr_svg_(self):
        project = self.project_shape()
        if project is not None:
            return project._repr_svg_()

    def _repr_html_(self):
        no_wrap_div = '<div style="white-space: nowrap">{}</div>'
        return no_wrap_div.format(self._repr_svg_())

    @lru_cache()
    def project_shape(self, projection=None):
        """Projection for a decent representation of the structure.

        By default, an equivalent projection is applied. Equivalent projections
        locally respect areas, which is convenient for the area item.
        """

        if self.shape is None:
            return None

        if projection is None:
            bounds = self.bounds
            projection = pyproj.Proj(
                proj="aea",  # equivalent projection
                lat1=bounds[1],
                lat2=bounds[3],
                lon1=bounds[0],
                lon2=bounds[2],
            )

        return transform(
            partial(
                pyproj.transform, pyproj.Proj(init="EPSG:4326"), projection
            ),
            self.shape,
        )

    # Nothing special made on plots because it really depends on the nature of
    # the geometry of the shape held in the structure.


class GeographyMixin(object):
    def compute_xy(self, projection=None):
        """Computes x and y columns from latitudes and longitudes.

        The source projection is WGS84 (EPSG 4326).
        The default destination projection is a Lambert Conformal Conical
        projection centered on the data inside the dataframe.

        For consistency reasons with pandas DataFrame, a new Traffic structure
        is returned.

        """
        if projection is None:
            projection = pyproj.Proj(
                proj="lcc",
                lat_1=self.data.latitude.min(),
                lat_2=self.data.latitude.max(),
                lat_0=self.data.latitude.mean(),
                lon_0=self.data.longitude.mean(),
            )

        x, y = pyproj.transform(
            pyproj.Proj(init="EPSG:4326"),
            projection,
            self.data.longitude.values,
            self.data.latitude.values,
        )

        return self.__class__(self.data.assign(x=x, y=y))



import warnings
from calendar import timegm
from datetime import datetime, timedelta
# from decimal import ROUND_DOWN, ROUND_UP, Decimal
from io import StringIO
from pathlib import Path
from typing import Dict, Iterable, Iterator, List, Optional, Set, Tuple, Union

import numpy as np

import pandas as pd
import pyproj
from cartopy.crs import PlateCarree
from scipy.interpolate import interp1d
from shapely.geometry import LineString, base

from ...core import Flight as FlightMixin
from ...core.mixins import DataFrameMixin
from ...core.time import time_or_delta, timelike, to_datetime
from ..sectors.airac import Sector


def time(int_: int) -> datetime:
    ts = timegm((2000 + int_ // 10000, int_ // 100 % 100, int_ % 100, 0, 0, 0))
    return datetime.fromtimestamp(ts)


def hour(int_: int) -> timedelta:
    return timedelta(
        hours=int_ // 10000, minutes=int_ // 100 % 100, seconds=int_ % 100
    )


class Flight(FlightMixin):

    def __init__(self, data: pd.DataFrame) -> None:
        super().__init__(data)
        self.interpolator: Dict = dict()

    @property
    def timestamp(self) -> Iterator[datetime]:
        if self.data.shape[0] == 0:
            return
        for _, s in self.data.iterrows():
            yield s.time1
        yield s.time2

    @property
    def aircraft(self) -> str:
        return self.data.iloc[0].aircraft

    @property
    def registration(self) -> None:
        return None

    @property
    def coords(self) -> Iterator[Tuple[float, float, float]]:
        if self.data.shape[0] == 0:
            return
        for _, s in self.data.iterrows():
            yield s.lon1, s.lat1, s.alt1
        yield s.lon2, s.lat2, s.alt2

    @property
    def linestring(self) -> LineString:
        return LineString(list(self.coords))

    @property
    def shape(self) -> LineString:
        return self.linestring

    def airborne(self):
        return self

    def interpolate(self, times, proj=PlateCarree()):
        """Interpolates a trajectory in time.  """
        if proj not in self.interpolator:
            self.interpolator[proj] = interp1d(
                np.stack(t.timestamp() for t in self.timestamp),
                proj.transform_points(
                    PlateCarree(), *np.stack(self.coords).T
                ).T,
            )
        return PlateCarree().transform_points(
            proj, *self.interpolator[proj](times)
        )

    def at(self, time: timelike, proj=PlateCarree()) -> np.ndarray:
        time = to_datetime(time)
        timearray: np.ndarray[datetime] = np.array([time.timestamp()])
        return self.interpolate(timearray, proj)

    def between(self, before: timelike, after: time_or_delta) -> "Flight":
        before = to_datetime(before)
        if isinstance(after, timedelta):
            after = before + after
        else:
            after = to_datetime(after)

        t: np.ndarray = np.stack(self.timestamp)
        index = np.where((before < t) & (t < after))

        new_data: np.ndarray = np.stack(self.coords)[index]
        time1: List[datetime] = [before, *t[index]]
        time2: List[datetime] = [*t[index], after]

        if before > t[0]:
            new_data = np.vstack([self.at(before), new_data])
        else:
            time1, time2 = time1[1:], time2[1:]
        if after < t[-1]:
            new_data = np.vstack([new_data, self.at(after)])
        else:
            time1, time2 = time1[:-1], time2[:-1]

        df: pd.DataFrame = (
            pd.DataFrame.from_records(
                np.c_[new_data[:-1, :], new_data[1:, :]],
                columns=["lon1", "lat1", "alt1", "lon2", "lat2", "alt2"],
            ).assign(
                time1=time1,
                time2=time2,
                origin=self.origin,
                destination=self.destination,
                aircraft=self.aircraft,
                flight_id=self.flight_id,
                callsign=self.callsign,
            )
        )

        return Flight(df)

    def clip(self, shape: base.BaseGeometry) -> "Flight":
        coords = np.stack(self.linestring.intersection(shape).coords)
        times = list(
            datetime.fromtimestamp(t)
            for t in np.stack(
                LineString(list(self.xy_time)).intersection(shape).coords
            )[:, 2]
        )

        df: pd.DataFrame = (
            pd.DataFrame.from_records(
                np.c_[coords[:-1, :], coords[1:, :]],
                columns=["lon1", "lat1", "alt1", "lon2", "lat2", "alt2"],
            ).assign(
                time1=times[:-1],
                time2=times[1:],
                origin=self.origin,
                destination=self.destination,
                aircraft=self.aircraft,
                flight_id=self.flight_id,
                callsign=self.callsign,
            )
        )
        return Flight(df)

    def clip_altitude(self, min_: int, max_: int) -> Iterator["Flight"]:
        def buffer_to_iter(proj, buffer):
            df = pd.DataFrame.from_records(buffer)

            df["lon1"], df["lat1"] = pyproj.transform(
                proj, pyproj.Proj(init="EPSG:4326"), df.x1.values, df.y1.values
            )

            df["lon2"], df["lat2"] = pyproj.transform(
                proj, pyproj.Proj(init="EPSG:4326"), df.x2.values, df.y2.values
            )

            yield df.drop(["x1", "x2", "y1", "y2"], axis=1)

        data = self.data.copy()

        proj = pyproj.Proj(
            proj="lcc",
            lat0=data.lat1.mean(),
            lon0=data.lon1.mean(),
            lat1=data.lat1.min(),
            lat2=data.lat1.max(),
        )

        data["x1"], data["y1"] = pyproj.transform(
            pyproj.Proj(init="EPSG:4326"),
            proj,
            data.lon1.values,
            data.lat1.values,
        )

        data["x2"], data["y2"] = pyproj.transform(
            pyproj.Proj(init="EPSG:4326"),
            proj,
            data.lon2.values,
            data.lat2.values,
        )

        buffer = []
        for (_, line) in data.iterrows():
            if (line.alt1 < max_ or line.alt2 < max_) and (
                line.alt1 > min_ or line.alt2 > min_
            ):

                if line.alt1 != line.alt2:
                    f_x = (line.x1 - line.x2) / (line.alt1 - line.alt2)
                    f_y = (line.y1 - line.y2) / (line.alt1 - line.alt2)

                if line.alt1 > max_:
                    line.x1 = line.x2 + (max_ - line.alt2) * f_x
                    line.y1 = line.y2 + (max_ - line.alt2) * f_y
                    line.alt1 = max_
                if line.alt1 < min_:
                    line.x1 = line.x2 + (min_ - line.alt2) * f_x
                    line.y1 = line.y2 + (min_ - line.alt2) * f_y
                    line.alt1 = min_
                if line.alt2 > max_:
                    line.x2 = line.x1 + (max_ - line.alt1) * f_x
                    line.y2 = line.y1 + (max_ - line.alt1) * f_y
                    line.alt2 = max_
                if line.alt2 < min_:
                    line.x2 = line.x1 + (min_ - line.alt1) * f_x
                    line.y2 = line.y1 + (min_ - line.alt1) * f_y
                    line.alt2 = min_

                buffer.append(line)

            else:
                if len(buffer) > 0:
                    yield from buffer_to_iter(proj, buffer)
                    buffer = []

        if len(buffer) > 0:
            yield from buffer_to_iter(proj, buffer)

    def resample(self):
        raise NotImplementedError("SO6 do not provide a resample method")


class SO6(DataFrameMixin):

    identifier = Union[int, str]

    def __getitem__(self, _id: identifier) -> Flight:
        if isinstance(_id, int):
            return Flight(self.data.groupby("flight_id").get_group(_id))
        if isinstance(_id, str):
            return Flight(self.data.groupby("callsign").get_group(_id))

    def __iter__(self) -> Iterator[Tuple[int, Flight]]:
        for flight_id, flight in self.data.groupby("flight_id"):
            yield flight_id, Flight(flight)

    def __len__(self) -> int:
        return len(self.flight_ids)

    def _ipython_key_completions_(self):
        return {*self.flight_ids, *self.callsigns}

    def get(self, callsign: str) -> Iterable[Tuple[int, Flight]]:
        all_flights = self.data.groupby("callsign").get_group(callsign)
        for flight_id, flight in all_flights.groupby("flight_id"):
            yield flight_id, Flight(flight)

    @property
    def callsigns(self) -> Set[str]:
        return set(self.data.callsign)

    @property
    def flight_ids(self) -> Set[int]:
        return set(self.data.flight_id)

    @classmethod
    def from_so6(self, filename: Union[str, StringIO]) -> "SO6":
        so6 = pd.read_csv(
            filename,
            sep=" ",
            header=-1,
            names=[
                "d1",
                "origin",
                "destination",
                "aircraft",
                "hour1",
                "hour2",
                "alt1",
                "alt2",
                "d2",
                "callsign",
                "date1",
                "date2",
                "lat1",
                "lon1",
                "lat2",
                "lon2",
                "flight_id",
                "d3",
                "d4",
                "d5",
            ],
        )

        so6 = so6.assign(
            lat1=so6.lat1 / 60,
            lat2=so6.lat2 / 60,
            lon1=so6.lon1 / 60,
            lon2=so6.lon2 / 60,
            alt1=so6.alt1 * 100,
            alt2=so6.alt2 * 100,
            time1=so6.date1.apply(time) + so6.hour1.apply(hour),
            time2=so6.date2.apply(time) + so6.hour2.apply(hour),
        )

        for col in (
            "d1",
            "d2",
            "d3",
            "d4",
            "d5",
            "date1",
            "date2",
            "hour1",
            "hour2",
        ):
            del so6[col]

        return SO6(so6)

    @classmethod
    def from_so6_7z(self, filename: str) -> "SO6":
        from libarchive.public import memory_reader

        with open(filename, "rb") as fh:
            with memory_reader(fh.read()) as entries:
                s = StringIO()
                for file in entries:
                    for block in file.get_blocks():
                        s.write(block.decode())
                s.seek(0)
                so6 = SO6.from_so6(s)
                s.close()
                return so6

    @classmethod
    def parse_file(cls, filename: str) -> Optional["SO6"]:
        warnings.warn("Use SO6.from_file(filename)", DeprecationWarning)
        return cls.from_file(filename)

    @classmethod
    def from_file(cls, filename: str) -> Optional["SO6"]:
        path = Path(filename)
        if path.suffixes == [".so6", ".7z"]:
            return SO6.from_so6_7z(filename)
        if path.suffixes == [".so6"]:
            return SO6.from_so6(filename)
        holder = DataFrameMixin.from_file(filename)
        return SO6(holder.data) if holder is not None else None

    def at(self, time: timelike) -> "SO6":
        time = to_datetime(time)
        return SO6(
            self.data[(self.data.time1 <= time) & (self.data.time2 > time)]
        )

    def between(self, before: timelike, after: time_or_delta) -> "SO6":
        before = to_datetime(before)
        if isinstance(after, timedelta):
            after = before + after
        else:
            after = to_datetime(after)
        return SO6(
            self.data[(self.data.time1 <= after) & (self.data.time2 >= before)]
        )

    def intersects(self, sector: Sector) -> "SO6":
        return SO6(
            self.data.groupby("flight_id").filter(
                lambda flight: Flight(flight).intersects(sector)
            )
        )

    def inside_bbox(self, bounds: Union[Sector, Tuple[float, ...]]) -> "SO6":

        if isinstance(bounds, Sector):
            bounds = bounds.flatten().bounds

        if isinstance(bounds, base.BaseGeometry):
            bounds = bounds.bounds

        west, south, east, north = bounds

        # Transform coords into intelligible floats
        # '-2.06 <= lon1 <= 4.50 & 42.36 <= lat1 <= 48.14', instead of
        #  (-2.066666603088379, 42.366943359375, 4.491666793823242,
        #   48.13333511352539)
        # dec = Decimal('0.00')
        # west = Decimal(west).quantize(dec, rounding=ROUND_DOWN)
        # east = Decimal(east).quantize(dec, rounding=ROUND_UP)
        # south = Decimal(south).quantize(dec, rounding=ROUND_DOWN)
        # north = Decimal(north).quantize(dec, rounding=ROUND_UP)

        # the numexpr query is 10% faster than the regular
        # data[data.lat1 >= ...] conjunctions of comparisons
        query = "{0} <= lon1 <= {2} & {1} <= lat1 <= {3}"
        query = query.format(west, south, east, north)

        data = self.data.query(query)

        callsigns: Set[str] = set(data.callsign)

        return SO6(
            self.data.groupby("flight_id").filter(
                lambda data: data.iloc[0].callsign in callsigns
            )
        )

    def select(self, query: Union["SO6", Iterable[str]]) -> "SO6":
        if isinstance(query, SO6):
            # not very natural, but why not...
            query = query.callsigns
        select = self.data.callsign.isin(query)
        return SO6(self.data[select])



try:
    from cartotools.crs import *  # noqa: F401 F403
    from cartotools.osm import location
except ImportError:
    # cartotools provides a few more basic projections
    from cartopy.crs import *  # noqa: F401 F403
    # Basic version of the complete cached requests included in cartotools
    from .location import location  # noqa: F401

from cartopy.feature import NaturalEarthFeature
from cartopy.mpl.geoaxes import GeoAxesSubplot

from ..core.mixins import ShapelyMixin

def countries(**kwargs):
    params = {'category': 'cultural',
              'name': 'admin_0_countries',
              'scale': '10m',
              'edgecolor': '#524c50',
              'facecolor': 'none',
              'alpha': .5,
              **kwargs}
    return NaturalEarthFeature(**params)


def rivers(**kwargs):
    params = {'category': 'physical',
              'name': 'rivers_lake_centerlines',
              'scale': '10m',
              'edgecolor': '#226666',
              'facecolor': 'none',
              'alpha': .5,
              **kwargs}
    return NaturalEarthFeature(**params)


def lakes(**kwargs):
    params = {'category': 'physical',
              'name': 'lakes',
              'scale': '10m',
              'edgecolor': '#226666',
              'facecolor': '#226666',
              'alpha': .2,
              **kwargs}
    return NaturalEarthFeature(**params)


def ocean(**kwargs):
    params = {'category': 'physical',
              'name': 'ocean',
              'scale': '10m',
              'edgecolor': '#226666',
              'facecolor': '#226666',
              'alpha': .2,
              **kwargs}
    return NaturalEarthFeature(**params)


def _set_default_extent(self):
    """Helper for a default extent limited to the projection boundaries."""
    west, south, east, north = self.projection.boundary.bounds
    self.set_extent((west, east, south, north), crs=self.projection)


GeoAxesSubplot.set_default_extent = _set_default_extent

def _set_extent(self, shape):
    if isinstance(shape, ShapelyMixin):
        return self._set_extent(shape.extent)
    self._set_extent(shape)

GeoAxesSubplot._set_extent = GeoAxesSubplot.set_extent
GeoAxesSubplot.set_extent = _set_extent



import logging
from functools import lru_cache
from pathlib import Path
from typing import Callable, Dict, Iterator, Optional, Set, Tuple, Union

import pandas as pd
import pyproj
from shapely.geometry import base

from .flight import Flight
from .mixins import DataFrameMixin, GeographyMixin


class Traffic(DataFrameMixin, GeographyMixin):

    _parse_extension: Dict[str, Callable[..., pd.DataFrame]] = dict()

    @classmethod
    def from_flights(cls, flights: Iterator[Flight]):
        return cls(pd.concat([f.data for f in flights]))

    @classmethod
    def from_file(
        cls, filename: Union[Path, str], **kwargs
    ) -> Optional['Traffic']:

        tentative = super().from_file(filename)
        if tentative is not None:
            return cls(tentative.data)

        path = Path(filename)
        method = cls._parse_extension.get("".join(path.suffixes), None)
        if method is None:
            logging.warn(f"{path.suffixes} extension is not supported")
            return None

        data = method(filename, **kwargs)
        if data is None:
            return None

        return cls(data)

    # --- Special methods ---

    def __add__(self, other) -> "Traffic":
        # useful for compatibility with sum() function
        if other == 0:
            return self
        return self.__class__(pd.concat([self.data, other.data]))

    def __radd__(self, other) -> "Traffic":
        return self + other

    def __getitem__(self, index: str) -> Optional[Flight]:

        if self.flight_ids is not None:
            data = self.data[self.data.flight_id == index]
            if data.shape[0] > 0:
                return Flight(data)

        # if no such index as flight_id or no flight_id column
        try:
            value16 = int(index, 16)  # noqa: F841 (unused value16)
            data = self.data[self.data.icao24 == index]
        except ValueError:
            data = self.data[self.data.callsign == index]

        if data.shape[0] > 0:
            return Flight(data)

        return None

    def _ipython_key_completions_(self) -> Set[str]:
        if self.flight_ids is not None:
            return self.flight_ids
        return {*self.aircraft, *self.callsigns}

    def __iter__(self):
        if self.flight_ids is not None:
            for _, df in self.data.groupby("flight_id"):
                yield Flight(df)
        else:
            for _, df in self.data.groupby("icao24"):
                yield from Flight(df).split()

    def __repr__(self):
        return self.stats().__repr__()

    def _repr_html_(self):
        stats = self.stats()
        shape = stats.shape[0]
        if shape > 10:
            # stylers are not efficient on big dataframes...
            stats = stats.head(10)
        styler = stats.style.bar(align="mid", color="#5fba7d")
        rep = f"<b>Traffic with {shape} identifiers</b>"
        return rep + styler._repr_html_()

    # --- Properties ---

    @property
    def start_time(self):
        return self.data.timestamp.min()

    @property
    def end_time(self):
        return self.data.timestamp.max()

    # https://github.com/python/mypy/issues/1362
    @property  # type: ignore
    @lru_cache()
    def callsigns(self) -> Set[str]:
        """Return only the most relevant callsigns"""
        sub = (
            self.data.query("callsign == callsign")
            .groupby(("callsign", "icao24"))
            .filter(lambda x: len(x) > 10)
        )
        return set(cs for cs in sub.callsign if len(cs) > 3 and " " not in cs)

    @property
    def aircraft(self) -> Set[str]:
        return set(self.data.icao24)

    @property
    def flight_ids(self) -> Optional[Set[str]]:
        if "flight_id" in self.data.columns:
            return set(self.data.flight_id)
        return None

    # --- Easy work ---

    @lru_cache()
    def stats(self) -> pd.DataFrame:
        """Statistics about flights contained in the structure.
        Useful for a meaningful representation.
        """
        key = ("icao24", "callsign") if self.flight_ids is None else "flight_id"
        return (
            self.data.groupby(key)[["timestamp"]]
            .count()
            .sort_values("timestamp", ascending=False)
            .rename(columns={"timestamp": "count"})
        )

    def plot(self, ax, **kwargs):
        params = {}
        if sum(1 for _ in zip(range(8), self)) == 8:
            params['color'] = '#aaaaaa'
            params['linewidth'] = 1
            params['alpha'] = .8
            kwargs = {**params, **kwargs}  # precedence of kwargs over params
        for flight in self:
            flight.plot(ax, **kwargs)

    # --- Real work ---

    def resample(self, rule="1s", kernel=(10, "m")):
        """Resamples all trajectories, flight by flight.

        `rule` defines the desired sample rate (default: 1s)
        `kernel` defines how to iter on flights (see `Flight.split`)
        """
        cumul = []
        for flight in self:
            for subflight in flight.split(*kernel):
                cumul.append(subflight.resample(rule))
        return self.__class__.from_flights(cumul)



import logging
from datetime import datetime, timedelta
from typing import Dict, Iterator, Optional, Set, Tuple, Union

import numpy as np

import pandas as pd
import pyproj
from shapely.geometry import LineString, base

from ..core.time import time_or_delta, timelike, to_datetime
from .mixins import DataFrameMixin, GeographyMixin, ShapelyMixin


def split(data: pd.DataFrame, value, unit) -> Iterator[pd.DataFrame]:
    diff = data.timestamp.diff().values
    if diff.max() > np.timedelta64(value, unit):
        yield from split(data.iloc[: diff.argmax()], value, unit)
        yield from split(data.iloc[diff.argmax() :], value, unit)
    else:
        yield data


class Flight(DataFrameMixin, ShapelyMixin, GeographyMixin):
    def __add__(self, other):
        # useful for compatibility with sum() function
        if other == 0:
            return self
        # keep import here to avoid recursion
        from .traffic import Traffic

        return Traffic.from_flights([self, other])

    def __radd__(self, other):
        return self + other

    def info_html(self) -> str:
        title = f"<b>Flight {self.callsign}</b>"
        if self.number is not None:
            title += f" / {self.number}"
        if self.flight_id is not None:
            title += f" ({self.flight_id})"

        title += "<ul>"
        title += f"<li><b>aircraft:</b> {self.aircraft}</li>"
        if self.origin is not None:
            title += f"<li><b>origin:</b> {self.origin} ({self.start})</li>"
        else:
            title += f"<li><b>origin:</b> {self.start}</li>"
        if self.destination is not None:
            title += f"<li><b>destination:</b> {self.destination} "
            title += f"({self.stop})</li>"
        else:
            title += f"<li><b>destination:</b> {self.stop}</li>"
        title += "</ul>"
        return title

    def _repr_html_(self):
        title = self.info_html()
        no_wrap_div = '<div style="white-space: nowrap">{}</div>'
        return title + no_wrap_div.format(self._repr_svg_())

    @property
    def timestamp(self) -> Iterator[datetime]:
        yield from self.data.timestamp

    @property
    def start(self) -> datetime:
        return min(self.timestamp)

    @property
    def stop(self) -> datetime:
        return max(self.timestamp)

    @property
    def callsign(self) -> Union[str, Set[str]]:
        tmp = set(self.data.callsign)
        if len(tmp) == 1:
            return tmp.pop()
        logging.warn("Several callsigns for one flight, consider splitting")
        return tmp

    @property
    def number(self) -> Optional[Union[str, Set[str]]]:
        if "number" not in self.data.columns:
            return None
        tmp = set(self.data.number)
        if len(tmp) == 1:
            return tmp.pop()
        logging.warn("Several numbers for one flight, consider splitting")
        return tmp

    @property
    def icao24(self) -> Union[str, Set[str]]:
        tmp = set(self.data.icao24)
        if len(tmp) == 1:
            return tmp.pop()
        logging.warn("Several icao24 for one flight, consider splitting")
        return tmp

    @property
    def flight_id(self) -> Optional[Union[str, Set[str]]]:
        if "flight_id" not in self.data.columns:
            return None
        tmp = set(self.data.flight_id)
        if len(tmp) == 1:
            return tmp.pop()
        logging.warn("Several ids for one flight, consider splitting")
        return tmp

    @property
    def origin(self) -> Optional[Union[str, Set[str]]]:
        if "origin" not in self.data.columns:
            return None
        tmp = set(self.data.origin)
        if len(tmp) == 1:
            return tmp.pop()
        logging.warn("Several origins for one flight, consider splitting")
        return tmp

    @property
    def destination(self) -> Optional[Union[str, Set[str]]]:
        if "destination" not in self.data.columns:
            return None
        tmp = set(self.data.destination)
        if len(tmp) == 1:
            return tmp.pop()
        logging.warn("Several destinations for one flight, consider splitting")
        return tmp

    @property
    def aircraft(self) -> Optional[str]:
        if not isinstance(self.icao24, str):
            return None
        from ..data import aircraft as acdb

        ac = acdb[self.icao24]
        if ac.shape[0] != 1:
            return self.icao24
        else:
            return f"{self.icao24} / {ac.iloc[0].regid} ({ac.iloc[0].mdl})"

    @property
    def registration(self) -> Optional[str]:
        from ..data import aircraft as acdb

        if not isinstance(self.icao24, str):
            return None
        ac = acdb[self.icao24]
        if ac.shape[0] != 1:
            return None
        return ac.iloc[0].regid

    @property
    def coords(self) -> Iterator[Tuple[float, float, float]]:
        data = self.data[self.data.longitude.notnull()]
        altitude = (
            "baro_altitude"
            if "baro_altitude" in self.data.columns
            else "altitude"
        )
        yield from zip(data["longitude"], data["latitude"], data[altitude])

    @property
    def xy_time(self) -> Iterator[Tuple[float, float, float]]:
        iterator = iter(zip(self.coords, self.timestamp))
        while True:
            next_ = next(iterator, None)
            if next_ is None:
                return
            coords, time = next_
            yield (coords[0], coords[1], time.timestamp())

    @property
    def linestring(self) -> Optional[LineString]:
        coords = list(self.coords)
        if len(coords) < 2:
            return None
        return LineString(coords)

    @property
    def shape(self) -> Optional[LineString]:
        return self.linestring

    def airborne(self) -> "Flight":
        altitude = (
            "baro_altitude"
            if "baro_altitude" in self.data.columns
            else "altitude"
        )
        return self.__class__(self.data[self.data[altitude].notnull()])

    # -- Interpolation and resampling --

    def split(self, value: int = 10, unit: str = "m") -> Iterator["Flight"]:
        for data in split(self.data, value, unit):
            yield self.__class__(data)

    def resample(self, rule: str = "1s") -> "Flight":
        data = (
            self.data.assign(start=self.start, stop=self.stop)
            .set_index("timestamp")
            .resample(rule)
            .interpolate()
            .reset_index()
            .fillna(method="pad")
        )
        return self.__class__(data)

    def at(self, time: timelike) -> pd.core.series.Series:
        index = to_datetime(time)
        return self.data.set_index("timestamp").loc[time]

    def between(self, before: timelike, after: time_or_delta) -> "Flight":
        before = to_datetime(before)
        if isinstance(after, timedelta):
            after = before + after
        else:
            after = to_datetime(after)

        t: np.ndarray = np.stack(self.timestamp)
        index = np.where((before < t) & (t < after))
        return self.__class__(self.data.iloc[index])

    # -- Geometry operations --

    def clip(
        self, shape: base.BaseGeometry
    ) -> Union[None, "Flight", Iterator["Flight"]]:


        linestring = LineString(list(self.airborne().xy_time))
        intersection = linestring.intersection(shape)

        if intersection.is_empty:
            return None

        if isinstance(intersection, LineString):
            times = list(
                datetime.fromtimestamp(t)
                for t in np.stack(intersection.coords)[:, 2]
            )
            return self.__class__(
                self.data[
                    (self.data.timestamp >= min(times))
                    & (self.data.timestamp <= max(times))
                ]
            )

        def _clip_generator():
            for segment in intersection:
                times = list(
                    datetime.fromtimestamp(t)
                    for t in np.stack(segment.coords)[:, 2]
                )
                yield self.__class__(
                    self.data[
                        (self.data.timestamp >= min(times))
                        & (self.data.timestamp <= max(times))
                    ]
                )

        return (leg for leg in _clip_generator())

    # -- Visualisation --

    def plot(self, ax, **kwargs):
        if "projection" in ax.__dict__ and "transform" not in kwargs:
            from cartopy.crs import PlateCarree

            kwargs["transform"] = PlateCarree()
        ax.plot(*self.shape.xy, **kwargs)



import sys

from .data import airports as data_airports
from .data import navaids as data_navaids
from .data import aircraft as data_aircraft


def get_airports(*args):
    for airport in data_airports.search(args[0]):
        print(airport)


def get_navaids(*args):
    for navaid in data_navaids.search(args[0]):
        print(navaid)


def get_aircraft(*args):
    if args[0] == "get":
        print(data_aircraft[args[1]])
    elif args[0] == "operator":
        print(data_aircraft.operator(" ".join(args[1:])))
    elif args[0] == "stats":
        print(data_aircraft.stats(" ".join(args[1:])))
    else:
        raise RuntimeError("Usage: traffic aircraft [get|operator|stats]")


cmd = {"airport": get_airports, "navaid": get_navaids, "aircraft": get_aircraft}


def main():
    command = sys.argv[1]
    cmd[command](*sys.argv[2:])



from datetime import datetime, timedelta
from numbers import Number
from typing import Iterator, Tuple, Union

import numpy as np

import maya
import pandas as pd

timelike = Union[str, Number, datetime, pd.Timestamp]
time_or_delta = Union[timelike, timedelta]
timetuple = Tuple[datetime, datetime, datetime, datetime]


def to_datetime(time: timelike) -> datetime:
    if isinstance(time, pd.Timestamp):
        time = time.to_pydatetime()
    if isinstance(time, str):
        time = maya.parse(time)  # type: ignore
    if isinstance(time, maya.core.MayaDT):  # type: ignore
        time = time.epoch
    if isinstance(time, Number):
        time = datetime.fromtimestamp(time)  # type: ignore
    return time  # type: ignore


def round_time(
    time: timelike, how: str = "before", by: timedelta = timedelta(hours=1)
) -> datetime:

    dt = to_datetime(time)

    round_to = by.total_seconds()
    seconds = (dt - dt.min).seconds

    if how == "after":
        rounding = (seconds + round_to) // round_to * round_to
    elif how == "before":
        rounding = seconds // round_to * round_to
    else:
        raise ValueError("parameter how must be `before` or `after`")

    return dt + timedelta(0, rounding - seconds, -dt.microsecond)


def split_times(
    before: datetime, after: datetime, by: timedelta = timedelta(hours=1)
) -> Iterator[timetuple]:

    before_hour = round_time(before, by=by)
    seq = np.arange(before_hour, after + by, by).astype(datetime)

    for bh, ah in zip(seq[:-1], seq[1:]):
        yield (before, after, bh, ah)



import json
from collections import UserDict
from pathlib import Path


class Cache(UserDict):

    def __init__(self, cachedir: Path) -> None:
        self.cachedir = cachedir
        if not self.cachedir.exists():
            self.cachedir.mkdir(parents=True)
        super().__init__()

    def __missing__(self, hashcode: str):
        filename = self.cachedir / f"{hashcode}.json"
        if filename.exists():
            response = json.loads(filename.read_text())
            # Do not overwrite the file!
            super().__setitem__(hashcode, response)
            return response

    def __setitem__(self, hashcode: str, data):
        super().__setitem__(hashcode, data)
        filename = self.cachedir / f"{hashcode}.json"
        with filename.open('w') as fh:
            fh.write(json.dumps(data))



# flake8: noqa
import numpy as np
import pandas as pd
import pyproj


def _douglas_peucker_rec(x: np.ndarray, y: np.ndarray,
                         mask: np.ndarray, tolerance: float) -> None:
    l = len(x)
    if l < 3: return

    v = np.array([[y[len(x)-1]-y[0]], [x[0] - x[len(x) -1]]])
    d = np.abs(np.dot(np.dstack([x[1:-1]-x[0], y[1:-1]-y[0]])[0],
                      v/np.sqrt(np.sum(v*v))))

    if np.max(d) < tolerance:
        mask[np.s_[1:l-1]] = 0
        return

    arg = np.argmax(d)
    _douglas_peucker_rec(x[:arg+2], y[:arg+2], mask[:arg+2], tolerance)
    _douglas_peucker_rec(x[arg+1:], y[arg+1:], mask[arg+1:], tolerance)


def _douglas_peucker_rec_3d(x: np.ndarray, y: np.ndarray, z: np.ndarray,
                            mask: np.ndarray, tolerance: float) -> None:
    l = len(x)
    if l < 3: return

    start = np.array([x[0], y[0], z[0]])
    end = np.array([x[-1], y[-1], z[-1]])
    point = np.dstack([x[1:], y[1:], z[1:]])[0] - start
    d = np.cross(point/np.linalg.norm(point), start-end)
    d = np.sqrt(np.sum(d*d, axis=1))

    if np.max(d) < tolerance:
        mask[np.s_[1:l-1]] = 0
        return

    arg = np.argmax(d)
    _douglas_peucker_rec_3d(x[:arg+2], y[:arg+2], z[:arg+2],
                            mask[:arg+2], tolerance)
    _douglas_peucker_rec_3d(x[arg+1:], y[arg+1:], z[arg+1:],
                            mask[arg+1:], tolerance)


def douglas_peucker(*args, df: pd.DataFrame=None, tolerance:float,
                    x='x', y='y', z=None, z_factor: float = 3.048,
                    lat=None, lon=None):
    """Ramer-Douglas-Peucker algorithm for 2D/3D trajectories.

    Simplify a trajectory by keeping the points further away from the straight
    line.

    Parameters:
        df        Optional                a Pandas dataframe
        tolerance float                   the threshold for cutting the
                                          trajectory
        z_factor  float                   for ft/m conversion (default 3.048)
                                            1km lateral, 100m vertical seems
                                            like a good ratio
        x, y, z   str or ndarray[float]   the column names if a dataframe is
                                          given, otherwise a series of float
        lat, lon  str or ndarray[float]   the column names if a dataframe is
                                          given, otherwise a series of float.
                                          x, y are built with a Lambert
                                          Conformal projection

        Note that lat, lon has precedence over x, y

    Returns:
        a np.array of booleans serving as a mask on the dataframe or
        on the numpy array

    See also: https://en.wikipedia.org/wiki/Ramer%E2%80%93Douglas%E2%80%93Peucker_algorithm

    """

    if df is None and (isinstance(x, str) or isinstance(y, str)):
        raise ValueError("Provide a dataframe if x and y are column names")
    if df is None and (isinstance(lon, str) or isinstance(lat, str)):
        raise ValueError("Provide a dataframe if lat and lon are column names")
    if tolerance < 0:
        raise ValueError("tolerance must be a positive float")

    if df is not None and isinstance(lat, str) and isinstance(lon, str):
        lat, lon = df[lat].values, df[lon].values
    if df is not None and lat is not None and lon is not None:
        lat, lon = np.array(lat), np.array(lon)
        x, y = pyproj.transform(
            pyproj.Proj(init='EPSG:4326'),
            pyproj.Proj(proj='lcc',
                        lat0=lat.mean(), lon0=lon.mean(),
                        lat1=lat.min(), lat2=lat.max(),
                        ), lon, lat)
    else:
        if df is not None:
            x, y = df[x].values, df[y].values
        x, y = np.array(x), np.array(y)

    if z is not None:
        if df is not None:
            z = df[z].values
        z = z_factor * np.array(z)

    mask = np.ones(len(x), dtype=bool)
    if z is None:
        _douglas_peucker_rec(x, y, mask, tolerance)
    else:
        _douglas_peucker_rec_3d(x, y, z, mask, tolerance)

    return mask



import re
from operator import itemgetter
from pathlib import Path
from typing import Optional, Set, Tuple, Union, cast

import pandas as pd
import requests
from shapely.geometry import LineString
from shapely.geometry.base import BaseGeometry
from shapely.ops import linemerge

from .navaid import Navaid  # type: ignore

__github_url = "https://raw.githubusercontent.com/"
base_url = __github_url + "ProfHoekstra/bluesky/master/data/navdata"

BoundsType = Union[BaseGeometry, Tuple[float, float, float, float]]


class Airways(object):

    cache: Optional[Path] = None

    def __init__(self, airways: Optional[pd.DataFrame]=None) -> None:
        if airways is None:
            if self.cache is not None and self.cache.exists():
                self.airways = pd.read_pickle(self.cache)
            else:
                self.initialize()
                # self.airways is set in above method
                self.airways = cast(pd.DataFrame, self.airways)
                if self.cache is not None:
                    self.airways.to_pickle(self.cache)
        else:
            self.airways = airways

    def initialize(self) -> None:
        c = requests.get(f"{base_url}/awy.dat")

        buffer = []
        for line in c.iter_lines():
            line = line.decode(encoding="ascii", errors="ignore").strip()
            if len(line) == 0 or line[0] == "#":
                continue
            fields = line.split()
            if len(fields) < 10:
                continue
            if not re.match("^[\d.]*$", fields[1]):
                continue

            for id_ in fields[-1].split("-"):
                buffer.append(
                    fields[:9] + [id_] +
                    [LineString([[float(fields[2]), float(fields[1])],
                                 [float(fields[5]), float(fields[4])]])])

        airways = pd.DataFrame.from_records(
            buffer, columns=['origin', 'fromlat', 'fromlon',
                             'destination', 'tolat', 'tolon',
                             'direction', 'low', 'up', 'id', 'linestring'])

        airways.low = airways.low.astype(int)
        airways.up = airways.up.astype(int)
        airways.fromlat = airways.fromlat.astype(float)
        airways.fromlon = airways.fromlon.astype(float)
        airways.tolat = airways.tolat.astype(float)
        airways.tolon = airways.tolon.astype(float)

        airways['bounds'] = airways.linestring.apply(lambda x: x.bounds)
        airways['west'] = airways.bounds.apply(itemgetter(0))
        airways['south'] = airways.bounds.apply(itemgetter(1))
        airways['east'] = airways.bounds.apply(itemgetter(2))
        airways['north'] = airways.bounds.apply(itemgetter(3))

        self.airways = airways

    def __getitem__(self, name: str) -> BaseGeometry:
        self.airways = cast(pd.DataFrame, self.airways)
        return linemerge(self.airways.groupby('id').get_group(name).linestring)

    def through(self, navaid: Union[str, Navaid],
                up_min: Optional[int] = None) -> Optional[Set[str]]:

        if isinstance(navaid, str):
            from .. import navaids
            _navaid = navaids[navaid]
            if _navaid is None:
                return None
            navaid = _navaid

        navaid = cast(Navaid, navaid)
        self.airways = cast(pd.DataFrame, self.airways)

        subset = self.airways[((self.airways.fromlat == navaid.lat) &
                               (self.airways.fromlon == navaid.lon)) |
                              ((self.airways.tolat == navaid.lat) &
                               (self.airways.tolon == navaid.lon))]

        if up_min is not None:
            subset = subset[subset.up >= up_min]

        return set(subset.id)

    def intersects(self, bounds: BoundsType,
                   up_min: Optional[int] = None) -> 'Airways':

        if isinstance(bounds, BaseGeometry):
            bounds = bounds.bounds

        west, south, east, north = bounds

        self.airways = cast(pd.DataFrame, self.airways)

        candidates = self.airways[
            ((self.airways.west >= west) | (self.airways.east >= west)) &
            ((self.airways.west <= east) | (self.airways.east <= east)) &
            ((self.airways.south >= south) | (self.airways.north >= south)) &
            ((self.airways.south <= north) | (self.airways.north <= north))]

        if up_min is not None:
            candidates = candidates[candidates.up >= up_min]

        return Airways(candidates)

    def plot(self, ax, **kwargs):

        if 'projection' in ax.__dict__:
            from cartopy.crs import PlateCarree
            kwargs['transform'] = PlateCarree()

        if 'color' not in kwargs:
            kwargs['color'] = '#aaaaaa'

        if 'linestyle' not in kwargs:
            kwargs['linestyle'] = ':'

        if 'lw' not in kwargs:
            kwargs['lw'] = .5

        for line in self.airways.linestring:
            ax.plot(*line.xy, **kwargs)



import io
import zipfile
from pathlib import Path
from typing import Optional

import pandas as pd
import requests


class Aircraft(object):

    cache: Optional[Path] = None

    def __init__(self):
        if self.cache is not None and self.cache.exists():
            self.aircraft = pd.read_pickle(self.cache)
        else:
            f = requests.get('https://junzisun.com/adb/download')
            with zipfile.ZipFile(io.BytesIO(f.content)) as zfile:
                with zfile.open('aircraft_db.csv', 'r') as dbfile:
                    self.aircraft = pd.read_csv(dbfile)
                    self.aircraft.regid = self.aircraft.regid.str.upper()
                    self.aircraft.mdl = self.aircraft.mdl.str.upper()
                    self.aircraft.to_pickle(self.cache)

    def __getitem__(self, name: str) -> pd.DataFrame:
        table = self.aircraft[(self.aircraft.icao == name.lower()) |
                              (self.aircraft.regid == name.upper())]
        return table

    def operator(self, name: str) -> pd.DataFrame:
        table = self.aircraft[self.aircraft.operator
                              .astype(str).str.contains(name)]
        return table

    def stats(self, name: str) -> pd.DataFrame:
        return self.operator(name).groupby(('mdl', 'type'))[['icao']].count()



import pickle
import re
from pathlib import Path
from typing import List, NamedTuple, Optional

import requests


class NavaidTuple(NamedTuple):

    id: str
    type: str
    lat: float
    lon: float
    alt: Optional[float]
    frequency: Optional[float]
    magnetic_variation: Optional[float]
    description: Optional[str]


class Navaid(NavaidTuple):
    def __repr__(self):
        if self.type == "FIX":
            return f"{self.id} ({self.type}): {self.lat} {self.lon}"
        else:
            return (
                f"{self.id} ({self.type}): {self.lat} {self.lon}"
                f" {self.alt:.0f}\n"
                f"{self.description if self.description is not None else ''}"
                f" {self.frequency}{'kHz' if self.type=='NDB' else 'MHz'}"
            )


__github_url = "https://raw.githubusercontent.com/"
base_url = __github_url + "ProfHoekstra/bluesky/master/data/navdata"


class NavaidParser(object):

    cache: Optional[Path] = None

    def __init__(self) -> None:
        if self.cache is not None and self.cache.exists():
            with self.cache.open("rb") as fh:
                self.navaids = pickle.load(fh)
        else:
            self.initialize()
            if self.cache is not None:
                with self.cache.open("wb") as fh:
                    pickle.dump(self.navaids, fh)

    def __getitem__(self, name: str) -> Optional[Navaid]:
        return next(
            (pt for pt in self.navaids if (pt.id == name.upper())), None
        )

    def search(self, name: str) -> List[Navaid]:
        return list(
            (
                pt
                for pt in self.navaids
                if (
                    pt.description is not None
                    and (re.match(name, pt.description, re.IGNORECASE))
                )
                or (pt.id == name.upper())
            )
        )

    def initialize(self):
        self.navaids = []
        c = requests.get(f"{base_url}/fix.dat")

        for line in c.iter_lines():

            line = line.decode(encoding="ascii", errors="ignore").strip()

            # Skip empty lines or comments
            if len(line) < 3 or line[0] == "#":
                continue

            # Start with valid 2 digit latitude -45. or 52.
            if not ((line[0] == "-" and line[3] == ".") or line[2] == "."):
                continue

            # Data line => Process fields of this record, separated by a comma
            # Example line:
            #  30.580372 -094.384169 FAREL
            fields = line.split()
            self.navaids.append(
                Navaid(
                    fields[2],
                    "FIX",
                    float(fields[0]),
                    float(fields[1]),
                    None,
                    None,
                    None,
                    None,
                )
            )

        c = requests.get(f"{base_url}/nav.dat")

        for line in c.iter_lines():

            line = line.decode(encoding="ascii", errors="ignore").strip()
            # Skip empty lines or comments
            if len(line) == 0 or line[0] == "#":
                continue

            # Data line => Process fields of this record, separated by a comma
            # Example lines:
            # 2  58.61466599  125.42666626 451   522  30  0.0 A   Aldan NDB
            # 3  31.26894444 -085.72630556 334 11120  40 -3.0 OZR CAIRNS VOR-DME
            # type    lat       lon        elev freq  ?   var id   desc
            #   0      1         2           3    4   5    6   7    8

            fields = line.split()

            # Valid line starst with integers
            if not fields[0].isdigit():
                continue  # Next line

            # Get code for type of navaid
            itype = int(fields[0])

            # Type names
            wptypedict = {
                2: "NDB",
                3: "VOR",
                4: "ILS",
                5: "LOC",
                6: "GS",
                7: "OM",
                8: "MM",
                9: "IM",
                12: "DME",
                13: "TACAN",
            }

            # Type code never larger than 20
            if itype not in list(wptypedict.keys()):
                continue  # Next line

            wptype = wptypedict[itype]

            # Select types to read
            if wptype not in ["NDB", "VOR", "DME", "TACAN"]:
                continue  # Next line

            # Find description
            try:
                idesc = line.index(fields[7]) + len(fields[7])
                description = line[idesc:].strip().upper()
            except Exception:
                description = None

            self.navaids.append(
                Navaid(
                    fields[7],
                    wptype,
                    float(fields[1]),
                    float(fields[2]),
                    float(fields[3]),
                    float(fields[4])
                    if wptype == "NDB"
                    else float(fields[4]) / 100,
                    float(fields[6]) if wptype in ["VOR", "NDB"] else None,
                    description,
                )
            )



import configparser
from pathlib import Path

from appdirs import user_cache_dir, user_config_dir

from .adsb.opensky import OpenSky
from .adsb import sqlite  # this could be a plugin...

from .sectors.airac import SectorParser
from .sectors.eurofirs import eurofirs
from .basic.aircraft import Aircraft
from .basic.airport import AirportParser
from .basic.airways import Airways
from .basic.navaid import NavaidParser

# Parse configuration and input specific parameters in below classes

__all__ = ['aircraft', 'airports', 'airways', 'navaids', 'sectors',
           'flightradar24', 'opensky', ]

config_dir = Path(user_config_dir("traffic"))
cache_dir = Path(user_cache_dir("traffic"))
config_file = config_dir / "traffic.conf"

if not config_dir.exists():
    config_dir.mkdir()
    with config_file.open('w') as fh:
        fh.write("""[global]
airac_path =
opensky_username =
opensky_password =
""")

if not cache_dir.exists():
    cache_dir.mkdir()

config = configparser.ConfigParser()
config.read(config_file.as_posix())

airac_path_str = config.get("global", "airac_path", fallback="")
if airac_path_str != "":
    SectorParser.airac_path = Path(airac_path_str)

Aircraft.cache = cache_dir / "aircraft.pkl"
AirportParser.cache = cache_dir / "airports.pkl"
Airways.cache = cache_dir / "airways.pkl"
NavaidParser.cache = cache_dir / "navaids.pkl"
SectorParser.cache_dir = cache_dir

aircraft = Aircraft()
airports = AirportParser()
airways = Airways()
navaids = NavaidParser()
airac = SectorParser(config_file)
sectors = airac  # deprecated?

opensky_username = config.get("global", "opensky_username", fallback="")
opensky_password = config.get("global", "opensky_password", fallback="")
opensky = OpenSky(opensky_username, opensky_password, cache_dir / "opensky")



import sqlite3
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import Dict, Union

import pandas as pd
from traffic.core import Traffic

""" TODO This parsing method is kept in a separate file as a stub for 'plugin'
for extending traffic structure.  """


def parse_sqlite(filename: Union[str, Path], with_ehs=False) -> pd.DataFrame:

    if isinstance(filename, Path):
        filename = filename.as_posix()

    connector = sqlite3.connect(filename)
    tables = ["surface", "positions", "vectors"]

    if with_ehs:
        tables += ["speed", "tracks", "meteo", "intention"]

    default_rename = {
        "lat": "latitude",
        "lon": "longitude",
        "alt": "altitude",
        "icao": "icao24",
        "speed": "ground_speed",
        "vertical": "vertical_rate",
    }

    renames: Dict[str, Dict[str, str]] = defaultdict(dict)
    renames["position"] = {"speed": "ground_speed"}
    renames["tracks"] = {"t_airspeed": "true_airspeed"}
    renames["vectors"] = {"heading": "track"}

    def posttreat(out: pd.DataFrame, more_rename=dict()) -> pd.DataFrame:
        df = out.rename(columns={**default_rename, **more_rename}).assign(
            timestamp=out.timestamp.apply(datetime.fromtimestamp),
            icao24=out.icao.apply(lambda x: hex(x)[2:]),
        )
        return df[df.callsign != "UNKNOWN"]

    return pd.concat(
        [
            posttreat(
                pd.read_sql_query(
                    f"select * from {table} order by timestamp", connector
                ),
                renames[table],
            )
            for table in tables
        ]
    ).sort_values("timestamp")


Traffic._parse_extension[".db"] = parse_sqlite



# flake8: noqa

from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt

from .cartopy import *

_traffic_style = '''
figure.figsize: 10, 10
figure.edgecolor: white
figure.facecolor: white

lines.linewidth: 1.5
lines.markeredgewidth: 0
lines.markersize: 10
lines.dash_capstyle: butt

legend.fancybox: True
font.size: 13

axes.prop_cycle: cycler('color', ['779ae3', 'e77074', '74c863', 'c2c55e', 'd27bcd', '60ceaf', 'd49146'])
axes.linewidth: 0
axes.titlesize: 22
axes.labelsize: 16

xtick.labelsize: 14
ytick.labelsize: 14
xtick.major.size: 0
xtick.minor.size: 0
ytick.major.size: 0
ytick.minor.size: 0

axes.grid: True
grid.alpha: 0.3
grid.linewidth: 0.5
grid.linestyle: --
grid.color: black

savefig.transparent: True
savefig.bbox: tight
savefig.format: png
'''

mpl_style_location = Path(f'{mpl.get_configdir()}/stylelib/traffic.mplstyle')
if not mpl_style_location.parent.is_dir():
    mpl_style_location.parent.mkdir(parents=True)

with mpl_style_location.open('w') as fh:
    fh.write(_traffic_style)

plt.style.reload_library()



import logging
from typing import Dict, List

import requests
from shapely.geometry import shape

from ..core.mixins import ShapelyMixin


def nominatim_request(query: str, **kwargs):

    params = {
        "format": "json",
        "limit": 1,
        "dedupe": 0,
        "polygon_geojson": 1,
        "q": query,
    }
    url = "https://nominatim.openstreetmap.org/search"
    response = requests.post(url, timeout=30, params=params, **kwargs)
    try:
        json = response.json()
    except Exception as e:
        logging.exception(e)  # type: ignore

    return json


class Nominatim(ShapelyMixin):

    def __init__(self, name: str, **kwargs) -> None:

        results: List[Dict] = nominatim_request(name, **kwargs)

        if len(results) == 0:
            raise ValueError(f"No '{name}' found on OpenStreetMap")

        self.json = results[0]
        self.shape = shape(self.json["geojson"])


def location(query: str, **kwargs):
    return Nominatim(query, **kwargs)



from setuptools import setup
from setuptools.extension import Extension
from Cython.Build import cythonize
import os.path

bresenham_path = os.path.join("traffic", "algorithms", "bresenham.pyx")
extensions = [Extension("traffic.algorithms.bresenham", [bresenham_path])]

setup(
    name="traffic",
    version=0.1,
    description="Tools for ATM",
    entry_points={"console_scripts": ["traffic=traffic.console:main"]},
    ext_modules=cythonize(extensions),
    packages=[
        "traffic",
        "traffic.core",
        "traffic.data",
        "traffic.data.adsb",
        "traffic.data.basic",
        "traffic.data.sectors",
        "traffic.data.so6",
        "traffic.algorithms",
        "traffic.drawing",
    ],
    package_data={"traffic.data.sectors": ["firs.json"]},
    install_requires=[
        "numpy",
        "scipy",
        "matplotlib",
        "pandas",
        "cython",
        "Cartopy",
        "Shapely",
        "requests",
        "maya",
        "appdirs",
        "paramiko",
        "tqdm",
        "cartotools==1.0",
    ],
    dependency_links=[
        "https://github.com/xoolive/cartotools.git#whl=cartotools-1.0"
    ],
    python_requires=">=3.6",
)



import logging
import pickle
import re
from functools import lru_cache
from pathlib import Path
from typing import NamedTuple, Optional

import requests
from cartopy.crs import PlateCarree

from ...core.mixins import ShapelyMixin


class AirportNamedTuple(NamedTuple):

    alt: int
    country: str
    iata: str
    icao: str
    lat: float
    lon: float
    name: str


class Airport(AirportNamedTuple, ShapelyMixin):

    def __repr__(self):
        return (
            f"{self.icao}/{self.iata}    {self.name.strip()} ({self.country})"
            f"\n\t{self.lat} {self.lon} altitude: {self.alt}"
        )

    def _repr_html_(self):
        title = f"<b>{self.name.strip()}</b> ({self.country}) "
        title += f"<code>{self.icao}/{self.iata}</code>"
        no_wrap_div = '<div style="white-space: nowrap">{}</div>'
        return title + no_wrap_div.format(self._repr_svg_())

    @lru_cache()
    def osm_request(self):
        from cartotools.osm import request, tags
        return request(
            (self.lon - .06, self.lat - .06, self.lon + .06, self.lat + .06),
            **tags.airport,
        )

    @property
    def shape(self):
        return self.osm_request().shape

    def plot(self, ax, **kwargs):
        params = {
            "edgecolor": "silver",
            "facecolor": "None",
            "crs": PlateCarree(),
            **kwargs,
        }
        ax.add_geometries(list(self.osm_request()), **params)


class AirportParser(object):

    cache: Optional[Path] = None

    def __init__(self):
        if self.cache is not None and self.cache.exists():
            with open(self.cache, "rb") as fh:
                self.airports = pickle.load(fh)
        else:
            c = requests.get("https://www.flightradar24.com/_json/airports.php",
                             headers={"user-agent": "Mozilla/5.0"})
            self.airports = list(Airport(**a) for a in c.json()["rows"])

            logging.info("Caching airport list from FlightRadar")
            if self.cache is not None:
                with open(self.cache, "wb") as fh:
                    pickle.dump(self.airports, fh)

    def __getitem__(self, name: str):
        return next(
            (
                a
                for a in self.airports
                if (a.iata == name.upper()) or (a.icao == name.upper())
            ),
            None,
        )

    def search(self, name: str):
        return list(
            (
                a
                for a in self.airports
                if (a.iata == name.upper())
                or (a.icao == name.upper())
                or (re.match(name, a.country, flags=re.IGNORECASE))
                or (re.match(name, a.name, flags=re.IGNORECASE))
            )
        )



from collections import defaultdict
from typing import (Any, Dict, Iterator, List, NamedTuple, Optional, Set,
                    Tuple, Union)

import numpy as np
from matplotlib.patches import Polygon as MplPolygon

from shapely.geometry import Polygon, base, mapping
from shapely.ops import cascaded_union

from . import Flight, Traffic
from .mixins import ShapelyMixin


class ExtrudedPolygon(NamedTuple):
    polygon: Polygon
    lower: float
    upper: float


class SectorInfo(NamedTuple):
    name: str
    type: Optional[str]


SectorList = List[ExtrudedPolygon]
components: Dict[str, Set[SectorInfo]] = defaultdict(set)


class Sector(ShapelyMixin):
    def __init__(
        self,
        name: str,
        elements: List[ExtrudedPolygon],
        type_: Optional[str] = None,
    ) -> None:
        self.elements: List[ExtrudedPolygon] = elements
        self.name: str = name
        self.type: Optional[str] = type_

    def flatten(self) -> Polygon:
        return cascaded_union([p.polygon for p in self])

    @property
    def shape(self):
        return self.flatten()

    def __getitem__(self, *args) -> ExtrudedPolygon:
        return self.elements.__getitem__(*args)

    def __add__(self, other: "Sector") -> "Sector":
        union = cascaded_union_with_alt(list(self) + list(other))
        return Sector(f"{self.name}+{other.name}", union)

    def __iter__(self) -> Iterator[ExtrudedPolygon]:
        return self.elements.__iter__()

    def _repr_html_(self):
        title = f"<b>{self.name} ({self.type})</b>"
        shapes = ""
        title += "<ul>"
        for polygon in self:
            title += f"<li>{polygon.lower}, {polygon.upper}</li>"
            shapes += polygon.polygon._repr_svg_()
        title += "</ul>"
        no_wrap_div = '<div style="white-space: nowrap">{}</div>'
        return title + no_wrap_div.format(shapes)

    def __repr__(self):
        return f"Sector {self.name} ({self.type})"

    def __str__(self):
        return f"""Sector {self.name} with {len(self.elements)} parts"""

    def annotate(self, ax, **kwargs):
        if "projection" in ax.__dict__:
            from cartopy.crs import PlateCarree

            kwargs["transform"] = PlateCarree()
        if "s" not in kwargs:
            kwargs["s"] = self.name
        ax.text(*np.array(self.centroid), **kwargs)

    def plot(self, ax, **kwargs):
        flat = self.flatten()
        if isinstance(flat, base.BaseMultipartGeometry):
            for poly in flat:
                # quick and dirty
                sub = Sector("", [ExtrudedPolygon(poly, 0, 0)])
                sub.plot(ax, **kwargs)
            return

        if "facecolor" not in kwargs:
            kwargs["facecolor"] = "None"
        if "edgecolor" not in kwargs:
            kwargs["edgecolor"] = "red"

        if "projection" in ax.__dict__:
            from cartopy.crs import PlateCarree
            ax.add_geometries([flat], crs=PlateCarree(), **kwargs)
        else:
            ax.add_patch(MplPolygon(list(flat.exterior.coords), **kwargs))

    @property
    def components(self) -> Set[SectorInfo]:
        return components[self.name]

    def decompose(self, extr_p):
        c = np.stack(extr_p.polygon.exterior.coords)
        alt = np.zeros(c.shape[0], dtype=float)

        alt[:] = min(extr_p.upper, 400) * 30.48
        upper_layer = np.c_[c, alt]
        yield Polygon(upper_layer)
        alt[:] = max(0, extr_p.lower) * 30.48
        lower_layer = np.c_[c, alt][::-1, :]
        yield Polygon(lower_layer)

        for i, j in zip(range(c.shape[0] - 1), range(c.shape[0], 1, -1)):
            yield Polygon(
                np.r_[
                    lower_layer[i:i+2, :],  # noqa: E226
                    upper_layer[j-2:j, :],  # noqa: E226
                ]
            )

    def export_json(self) -> Dict[str, Any]:
        export: Dict[str, Any] = {"name": self.name, "type": self.type}
        shapes = []
        for p in self:
            shapes.append(
                {
                    "upper": p.upper,
                    "lower": p.lower,
                    "polygon": mapping(p.polygon),
                }
            )
        export["shapes"] = shapes
        return export


def cascaded_union_with_alt(polyalt: SectorList) -> SectorList:
    altitudes = set(alt for _, *low_up in polyalt for alt in low_up)
    slices = sorted(altitudes)
    if len(slices) == 1 and slices[0] is None:
        simple_union = cascaded_union([p for p, *_ in polyalt])
        return [ExtrudedPolygon(simple_union, float("-inf"), float("inf"))]
    results: List[ExtrudedPolygon] = []
    for low, up in zip(slices, slices[1:]):
        matched_poly = [
            p
            for (p, low_, up_) in polyalt
            if low_ <= low <= up_ and low_ <= up <= up_
        ]
        new_poly = ExtrudedPolygon(cascaded_union(matched_poly), low, up)
        if len(results) > 0 and new_poly.polygon.equals(results[-1].polygon):
            merged = ExtrudedPolygon(new_poly.polygon, results[-1].lower, up)
            results[-1] = merged
        else:
            results.append(new_poly)
    return results


# -- Methods below are placed here because of possible circular imports --


def _traffic_inside_bbox(
    traffic: Traffic, bounds: Union[Sector, Tuple[float, ...]]
) -> Traffic:

    if isinstance(bounds, Sector):
        bounds = bounds.flatten().bounds

    if isinstance(bounds, base.BaseGeometry):
        bounds = bounds.bounds

    west, south, east, north = bounds

    query = "{0} <= longitude <= {2} & {1} <= latitude <= {3}"
    query = query.format(west, south, east, north)

    data = traffic.data.query(query)

    return traffic.__class__(data)


def _traffic_intersects(traffic: Traffic, sector: Sector) -> Traffic:
    return Traffic.from_flights(
        flight for flight in traffic if flight.intersects(sector)
    )


def _flight_intersects(flight: Flight, sector: Sector) -> bool:
    for layer in sector:
        linestring = flight.airborne().linestring
        if linestring is None:
            return False
        ix = linestring.intersection(layer.polygon)
        if not ix.is_empty:
            if isinstance(ix, base.BaseMultipartGeometry):
                for part in ix:
                    if any(
                        100 * layer.lower < x[2] < 100 * layer.upper
                        for x in part.coords
                    ):
                        return True
            else:
                if any(
                    100 * layer.lower < x[2] < 100 * layer.upper
                    for x in ix.coords
                ):
                    return True
    return False


# -- The ugly monkey-patching --

setattr(Traffic, "inside_bbox", _traffic_inside_bbox)
setattr(Traffic, "intersects", _traffic_intersects)
setattr(Flight, "intersects", _flight_intersects)
